@@ -1,4 +1,3 @@
-require 'open3'
 module Redwood
 
 class ThreadViewMode < LineCursorMode
@@ -11,7 +10,7 @@ class ThreadViewMode < LineCursorMode
     attr_accessor :state
   end
 
-  DATE_FORMAT = "%B %e %Y %l:%M%P"
+  DATE_FORMAT = "%B %e %Y %l:%M%p"
   INDENT_SPACES = 2 # how many spaces to indent child messages
 
   HookManager.register "detailed-headers", <<EOS
@@ -24,6 +23,18 @@ Return value:
   None. The variable 'headers' should be modified in place.
 EOS
 
+  HookManager.register "bounce-command", <<EOS
+Determines the command used to bounce a message.
+Variables:
+      from: The From header of the message being bounced
+            (eg: likely _not_ your address).
+        to: The addresses you asked the message to be bounced to as an array.
+Return value:
+  A string representing the command to pipe the mail into.  This
+  should include the entire command except for the destination addresses,
+  which will be appended by sup.
+EOS
+
   register_keymap do |k|
     k.add :toggle_detailed_header, "Toggle detailed header", 'h'
     k.add :show_header, "Show full message header", 'H'
@@ -41,16 +52,23 @@ EOS
     k.add :toggle_new, "Toggle unread/read status of message", 'N'
 #    k.add :collapse_non_new_messages, "Collapse all but unread messages", 'N'
     k.add :reply, "Reply to a message", 'r'
+    k.add :reply_all, "Reply to all participants of this message", 'G'
     k.add :forward, "Forward a message or attachment", 'f'
+    k.add :bounce, "Bounce message to other recipient(s)", '!'
     k.add :alias, "Edit alias/nickname for a person", 'i'
     k.add :edit_as_new, "Edit message as new", 'D'
     k.add :save_to_disk, "Save message/attachment to disk", 's'
+    k.add :save_all_to_disk, "Save all attachments to disk", 'A'
     k.add :search, "Search for messages from particular people", 'S'
     k.add :compose, "Compose message to person", 'm'
     k.add :subscribe_to_list, "Subscribe to/unsubscribe from mailing list", "("
     k.add :unsubscribe_from_list, "Subscribe to/unsubscribe from mailing list", ")"
     k.add :pipe_message, "Pipe message or attachment to a shell command", '|'
 
+    k.add :archive_and_next, "Archive this thread, kill buffer, and view next", 'a'
+    k.add :delete_and_next, "Delete this thread, kill buffer, and view next", 'd'
+    k.add :toggle_wrap, "Toggle wrapping of text", 'w'
+
     k.add_multi "(a)rchive/(d)elete/mark as (s)pam/mark as u(N)read:", '.' do |kk|
       kk.add :archive_and_kill, "Archive this thread and kill buffer", 'a'
       kk.add :delete_and_kill, "Delete this thread and kill buffer", 'd'
@@ -111,11 +129,19 @@ EOS
       end
     end
 
+    @wrap = true
+
     @layout[latest].state = :open if @layout[latest].state == :closed
     @layout[earliest].state = :detailed if earliest.has_label?(:unread) || @thread.size == 1
 
     @thread.remove_label :unread
+    Index.save_thread @thread
+  end
+
+  def toggle_wrap
+    @wrap = !@wrap
     regen_text
+    buffer.mark_dirty if buffer
   end
 
   def draw_line ln, opts={}
@@ -128,17 +154,25 @@ EOS
   def lines; @text.length; end
   def [] i; @text[i]; end
 
+  ## a little hacky---since regen_text can depend on buffer features like the
+  ## content_width, we don't call it in the constructor, and instead call it
+  ## here, which is set before we're responsible for drawing ourself.
+  def buffer= b
+    super
+    regen_text
+  end
+
   def show_header
     m = @message_lines[curpos] or return
     BufferManager.spawn_unless_exists("Full header for #{m.id}") do
-      TextMode.new m.raw_header
+      TextMode.new m.raw_header.ascii
     end
   end
 
   def show_message
     m = @message_lines[curpos] or return
     BufferManager.spawn_unless_exists("Raw message for #{m.id}") do
-      TextMode.new m.raw_message
+      TextMode.new m.raw_message.ascii
     end
   end
 
@@ -148,16 +182,18 @@ EOS
     update
   end
 
-  def reply
+  def reply type_arg=nil
     m = @message_lines[curpos] or return
-    mode = ReplyMode.new m
+    mode = ReplyMode.new m, type_arg
     BufferManager.spawn "Reply to #{m.subj}", mode
   end
 
+  def reply_all; reply :all; end
+
   def subscribe_to_list
     m = @message_lines[curpos] or return
-    if m.list_subscribe && m.list_subscribe =~ /<mailto:(.*?)\?(subject=(.*?))>/
-      ComposeMode.spawn_nicely :from => AccountManager.account_for(m.recipient_email), :to => [Person.from_address($1)], :subj => $3
+    if m.list_subscribe && m.list_subscribe =~ /<mailto:(.*?)(\?subject=(.*?))?>/
+      ComposeMode.spawn_nicely :from => AccountManager.account_for(m.recipient_email), :to => [Person.from_address($1)], :subj => ($3 || "subscribe")
     else
       BufferManager.flash "Can't find List-Subscribe header for this message."
     end
@@ -165,8 +201,8 @@ EOS
 
   def unsubscribe_from_list
     m = @message_lines[curpos] or return
-    if m.list_unsubscribe && m.list_unsubscribe =~ /<mailto:(.*?)\?(subject=(.*?))>/
-      ComposeMode.spawn_nicely :from => AccountManager.account_for(m.recipient_email), :to => [Person.from_address($1)], :subj => $3
+    if m.list_unsubscribe && m.list_unsubscribe =~ /<mailto:(.*?)(\?subject=(.*?))?>/
+      ComposeMode.spawn_nicely :from => AccountManager.account_for(m.recipient_email), :to => [Person.from_address($1)], :subj => ($3 || "unsubscribe")
     else
       BufferManager.flash "Can't find List-Unsubscribe header for this message."
     end
@@ -180,6 +216,33 @@ EOS
     end
   end
 
+  def bounce
+    m = @message_lines[curpos] or return
+    to = BufferManager.ask_for_contacts(:people, "Bounce To: ") or return
+
+    defcmd = AccountManager.default_account.bounce_sendmail
+
+    cmd = case (hookcmd = HookManager.run "bounce-command", :from => m.from, :to => to)
+          when nil, /^$/ then defcmd
+          else hookcmd
+          end + ' ' + to.map { |t| t.email }.join(' ')
+
+    bt = to.size > 1 ? "#{to.size} recipients" : to.to_s
+
+    if BufferManager.ask_yes_or_no "Really bounce to #{bt}?"
+      debug "bounce command: #{cmd}"
+      begin
+        IO.popen(cmd, 'w') do |sm|
+          sm.puts m.raw_message
+        end
+        raise SendmailCommandFailed, "Couldn't execute #{cmd}" unless $? == 0
+      rescue SystemCallError, SendmailCommandFailed => e
+        warn "problem sending mail: #{e.message}"
+        BufferManager.flash "Problem sending mail: #{e.message}"
+      end
+    end
+  end
+
   include CanAliasContacts
   def alias
     p = @person_lines[curpos] or return
@@ -204,14 +267,21 @@ EOS
   end    
 
   def edit_labels
-    reserved_labels = @thread.labels.select { |l| LabelManager::RESERVED_LABELS.include? l }
+    old_labels = @thread.labels
+    reserved_labels = old_labels.select { |l| LabelManager::RESERVED_LABELS.include? l }
     new_labels = BufferManager.ask_for_labels :label, "Labels for thread: ", @thread.labels
 
     return unless new_labels
-    @thread.labels = (reserved_labels + new_labels).uniq
+    @thread.labels = Set.new(reserved_labels) + new_labels
     new_labels.each { |l| LabelManager << l }
     update
     UpdateManager.relay self, :labeled, @thread.first
+    Index.save_thread @thread
+    UndoManager.register "labeling thread" do
+      @thread.labels = old_labels
+      Index.save_thread @thread
+      UpdateManager.relay self, :labeled, @thread.first
+    end
   end
 
   def toggle_starred
@@ -234,6 +304,7 @@ EOS
     ## star to the display
     update
     UpdateManager.relay self, :single_message_labeled, m
+    Index.save_thread @thread
   end
 
   ## called when someone presses enter when the cursor is highlighting
@@ -276,8 +347,10 @@ EOS
     chunk = @chunk_lines[curpos] or return
     case chunk
     when Chunk::Attachment
-      default_dir = File.join(($config[:default_attachment_save_dir] || "."), chunk.filename)
-      fn = BufferManager.ask_for_filename :filename, "Save attachment to file: ", default_dir
+      default_dir = $config[:default_attachment_save_dir]
+      default_dir = ENV["HOME"] if default_dir.nil? || default_dir.empty?
+      default_fn = File.expand_path File.join(default_dir, chunk.filename)
+      fn = BufferManager.ask_for_filename :filename, "Save attachment to file: ", default_fn
       save_to_file(fn) { |f| f.print chunk.raw_content } if fn
     else
       m = @message_lines[curpos]
@@ -289,6 +362,32 @@ EOS
     end
   end
 
+  def save_all_to_disk
+    m = @message_lines[curpos] or return
+    default_dir = ($config[:default_attachment_save_dir] || ".")
+    folder = BufferManager.ask_for_filename :filename, "Save all attachments to folder: ", default_dir, true
+    return unless folder
+
+    num = 0
+    num_errors = 0
+    m.chunks.each do |chunk|
+      next unless chunk.is_a?(Chunk::Attachment)
+      fn = File.join(folder, chunk.filename)
+      num_errors += 1 unless save_to_file(fn, false) { |f| f.print chunk.raw_content }
+      num += 1
+    end
+
+    if num == 0
+      BufferManager.flash "Didn't find any attachments!"
+    else
+      if num_errors == 0
+        BufferManager.flash "Wrote #{num.pluralize 'attachment'} to #{folder}."
+      else
+        BufferManager.flash "Wrote #{(num - num_errors).pluralize 'attachment'} to #{folder}; couldn't write #{num_errors} of them (see log)."
+      end
+    end
+  end
+
   def edit_draft
     m = @message_lines[curpos] or return
     if m.is_draft?
@@ -313,16 +412,16 @@ EOS
     end
   end
 
-  def jump_to_first_open loose_alignment=false
+  def jump_to_first_open
     m = @message_lines[0] or return
     if @layout[m].state != :closed
-      jump_to_message m, loose_alignment
+      jump_to_message m#, true
     else
-      jump_to_next_open loose_alignment
+      jump_to_next_open #true
     end
   end
 
-  def jump_to_next_open loose_alignment=false
+  def jump_to_next_open force_alignment=nil
     return continue_search_in_buffer if in_search? # hack: allow 'n' to apply to both operations
     m = (curpos ... @message_lines.length).argfind { |i| @message_lines[i] }
     return unless m
@@ -330,15 +429,15 @@ EOS
       break if @layout[nextm].state != :closed
       m = nextm
     end
-    jump_to_message nextm, loose_alignment if nextm
+    jump_to_message nextm, force_alignment if nextm
   end
 
   def align_current_message
     m = @message_lines[curpos] or return
-    jump_to_message m
+    jump_to_message m, true
   end
 
-  def jump_to_prev_open loose_alignment=false
+  def jump_to_prev_open
     m = (0 .. curpos).to_a.reverse.argfind { |i| @message_lines[i] } # bah, .to_a
     return unless m
     ## jump to the top of the current message if we're in the body;
@@ -350,38 +449,33 @@ EOS
         break if @layout[prevm].state != :closed
         m = prevm
       end
-      jump_to_message prevm, loose_alignment if prevm
+      jump_to_message prevm if prevm
     else
-      jump_to_message m, loose_alignment
+      jump_to_message m
     end
   end
 
-  IDEAL_TOP_CONTEXT = 3 # try and give 3 rows of top context
-  IDEAL_LEFT_CONTEXT = 4 # try and give 4 columns of left context
-  def jump_to_message m, loose_alignment=false
+  def jump_to_message m, force_alignment=false
     l = @layout[m]
-    left = l.depth * INDENT_SPACES
-    right = left + l.width
 
-    ## jump to the top line
-    if loose_alignment
-      jump_to_line [l.top - IDEAL_TOP_CONTEXT, 0].max # give 3 lines of top context
-    else
-      jump_to_line l.top
+    ## boundaries of the message
+    message_left = l.depth * INDENT_SPACES
+    message_right = message_left + l.width
+
+    ## calculate leftmost colum
+    left = if force_alignment # force mode: align exactly
+      message_left
+    else # regular: minimize cursor movement
+      ## leftmost and rightmost are boundaries of all valid left-column
+      ## alignments.
+      leftmost = [message_left, message_right - buffer.content_width + 1].min
+      rightmost = message_left
+      leftcol.clamp(leftmost, rightmost)
     end
 
-    ## jump to the left column
-    ideal_left = left +
-      if loose_alignment
-        -IDEAL_LEFT_CONTEXT + (l.width - buffer.content_width + IDEAL_LEFT_CONTEXT + 1).clamp(0, IDEAL_LEFT_CONTEXT)
-      else
-        0
-      end
-
-    jump_to_col [ideal_left, 0].max
-
-    ## either way, move the cursor to the first line
-    set_cursor_pos l.top
+    jump_to_line l.top    # move vertically
+    jump_to_col left      # move horizontally
+    set_cursor_pos l.top  # set cursor pos
   end
 
   def expand_all_messages
@@ -431,6 +525,12 @@ EOS
     dispatch op do
       @thread.remove_label :inbox
       UpdateManager.relay self, :archived, @thread.first
+      Index.save_thread @thread
+      UndoManager.register "archiving 1 thread" do
+        @thread.apply_label :inbox
+        Index.save_thread @thread
+        UpdateManager.relay self, :unarchived, @thread.first
+      end
     end
   end
 
@@ -438,6 +538,12 @@ EOS
     dispatch op do
       @thread.apply_label :spam
       UpdateManager.relay self, :spammed, @thread.first
+      Index.save_thread @thread
+      UndoManager.register "marking 1 thread as spam" do
+        @thread.remove_label :spam
+        Index.save_thread @thread
+        UpdateManager.relay self, :unspammed, @thread.first
+      end
     end
   end
 
@@ -445,6 +551,12 @@ EOS
     dispatch op do
       @thread.apply_label :deleted
       UpdateManager.relay self, :deleted, @thread.first
+      Index.save_thread @thread
+      UndoManager.register "deleting 1 thread" do
+        @thread.remove_label :deleted
+        Index.save_thread @thread
+        UpdateManager.relay self, :undeleted, @thread.first
+      end
     end
   end
 
@@ -452,6 +564,7 @@ EOS
     dispatch op do
       @thread.apply_label :unread
       UpdateManager.relay self, :unread, @thread.first
+      Index.save_thread @thread
     end
   end
 
@@ -500,7 +613,7 @@ EOS
     end
 
     if output
-      BufferManager.spawn "Output of '#{command}'", TextMode.new(output)
+      BufferManager.spawn "Output of '#{command}'", TextMode.new(output.ascii)
     else
       BufferManager.flash "'#{command}' done!"
     end
@@ -677,7 +790,12 @@ private
     else
       raise "Bad chunk: #{chunk.inspect}" unless chunk.respond_to?(:inlineable?) ## debugging
       if chunk.inlineable?
-        chunk.lines.map { |line| [[chunk.color, "#{prefix}#{line}"]] }
+        lines = chunk.lines
+        if @wrap
+          width = buffer.content_width
+          lines = lines.map { |l| l.chomp.wrap width }.flatten
+        end
+        lines.map { |line| [[chunk.color, "#{prefix}#{line}"]] }
       elsif chunk.expandable?
         case state
         when :closed
@@ -697,7 +815,7 @@ private
     BufferManager.erase_flash
     BufferManager.completely_redraw_screen
     unless success
-      BufferManager.spawn "Attachment: #{chunk.filename}", TextMode.new(chunk.to_s, chunk.filename)
+      BufferManager.spawn "Attachment: #{chunk.filename}", TextMode.new(chunk.to_s.ascii, chunk.filename)
       BufferManager.flash "Couldn't execute view command, viewing as text."
     end
   end
